from xappt.plugins.tools.examples.data_types import DataTypes
from xappt.plugins.tools.examples.modify_choices import ModifyChoices
from xappt.plugins.tools.examples.tool_chaining import ToolChaining



import enum
import os
import shlex
import signal
import subprocess
import warnings

from collections import namedtuple
from queue import Queue
from threading import Thread
from typing import List, Sequence, Union


CommandResult = namedtuple("CommandResult", ["result", "stdout", "stderr"])


def io_fn_default(_: str):
    """ Default subprocess io callback. """
    pass


def kill_pid(pid: int):
    try:
        os.kill(pid, signal.SIGTERM)
    except PermissionError:
        pass


class CommandRunnerState(enum.Enum):
    IDLE = 0
    RUNNING = 1
    ABORTED = 2


class PipeMonitor(Thread):
    def __init__(self, fd, queue):
        super().__init__()
        self._fd = fd
        self._queue = queue

    def run(self):
        for line in iter(self._fd.readline, ''):
            self._queue.put(line)

    def eof(self):
        return not self.is_alive() and self._queue.empty()


class CommandRunner(object):
    """ A wrapper around `subprocess.Popen` with basic environment manipulation.
    The results of a subprocess operation will be returned as a `CommandResult`
    object, which includes the return code, and stdout/stderr. Note that
    stdout/stderr will only be populated if the command was run silently.

    >>> c = CommandRunner(env={})
    >>> c.env_var_set("TEST", "1234")
    >>> c.env['TEST']
    '1234'
    >>> if os.name == "nt":
    ...     command = "set"
    ... elif os.name == "posix":
    ...     command = "printenv"
    >>> result = c.run(command, silent=True)
    >>> result.result == 0
    True
    >>> result.stdout.count("TEST=1234")
    1

    """
    def __init__(self, **kwargs):
        self.cwd = str(kwargs.get('cwd', os.getcwd()))
        self.env = kwargs.get('env', os.environ.copy())
        self._state = CommandRunnerState.IDLE

    def _split_path_var(self, key: str) -> List[str]:
        values = self.env.get(key, "").split(os.pathsep)
        return [v for v in values if len(v.strip())]

    def env_path_append(self, key: str, value: str):
        values = self._split_path_var(key)
        values.append(value)
        self.env[key] = os.pathsep.join(values)

    def env_path_prepend(self, key: str, value: str):
        values = self._split_path_var(key)
        values.insert(0, value)
        self.env[key] = os.pathsep.join(values)

    def env_var_add(self, key: str, value: str):
        """ This method was renamed to `env_var_set` to make it clear that
        `value` will replace anything that might already be set for `key`.
        The term "add" didn't really convey this. """
        warnings.warn("Call to deprecated function `env_var_add`. "
                      "Use `env_var_set` instead.", DeprecationWarning)
        self.env[key] = value

    def env_var_set(self, key: str, value: str):
        self.env[key] = value

    def env_var_remove(self, key: str):
        try:
            del self.env[key]
        except KeyError:
            pass

    def abort(self):
        self._state = CommandRunnerState.ABORTED

    @property
    def running(self):
        return self._state == CommandRunnerState.RUNNING

    def run(self, command: Union[bytes, str, Sequence], **kwargs) -> CommandResult:
        self._state = CommandRunnerState.RUNNING
        env = kwargs.get('env', self.env)
        shell = kwargs.get('shell', False)

        subprocess_args = {
            'cwd': str(kwargs.get('cwd') or self.cwd),
            'env': env,
            'shell': shell,
            'universal_newlines': True,
            'encoding': kwargs.get('encoding', 'utf8'),
        }

        if 'silent' in kwargs:
            warnings.warn("Using deprecated keyword argument `silent`. "
                          "Use `capture_output` instead.", DeprecationWarning)

        capture_output = kwargs.get('capture_output', True) or kwargs.get('silent', True)
        if capture_output:
            subprocess_args['stdout'] = subprocess.PIPE
            subprocess_args['stderr'] = subprocess.PIPE
        else:
            subprocess_args['stdout'] = kwargs.get('stdout')
            subprocess_args['stderr'] = kwargs.get('stderr')

        proc = subprocess.Popen(command, **subprocess_args)

        if capture_output:
            stdout_fn = kwargs.get('stdout_fn', io_fn_default)
            stderr_fn = kwargs.get('stderr_fn', io_fn_default)
            stdout = []
            stderr = []

            q_out = Queue()
            q_err = Queue()
            while proc.poll() is None:
                t_out = PipeMonitor(proc.stdout, q_out)
                t_err = PipeMonitor(proc.stderr, q_err)
                t_out.start()
                t_err.start()
                while not t_out.eof() or not t_err.eof():
                    while not q_out.empty():
                        line = q_out.get().rstrip()
                        stdout.append(line)
                        stdout_fn(line)
                    while not q_err.empty():
                        line = q_err.get().rstrip()
                        stderr.append(line)
                        stderr_fn(line)
                    if self._state == CommandRunnerState.ABORTED:
                        kill_pid(proc.pid)
                        break
                t_out.join()
                t_err.join()
            proc.stdout.close()
            proc.stderr.close()
            result = proc.returncode
            self._state = CommandRunnerState.IDLE
            return CommandResult(result, "\n".join(stdout), "\n".join(stderr))
        else:
            proc.communicate()
            self._state = CommandRunnerState.IDLE
            return CommandResult(proc.returncode, None, None)

    @staticmethod
    def command_sequence_to_string(command_seq: Sequence):
        if os.name == "nt":
            return subprocess.list2cmdline(command_seq)
        else:
            return ' '.join(shlex.quote(arg) for arg in command_seq)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from xappt.utilities import git_tools
from xappt.utilities.command_runner import CommandRunner, CommandResult
from xappt.utilities.find_python import find_python
from xappt.utilities.path import *
from xappt.utilities.humanize import *
from xappt.utilities import setup_helpers



from __future__ import annotations
import abc

from typing import Any, Optional, Sequence, Type, TYPE_CHECKING, Union

import xappt.managers.plugin_manager
from xappt.utilities.command_runner import CommandRunner

from xappt.models.plugins.base import BasePlugin
from xappt.models.parameter.model import Callback
if TYPE_CHECKING:
    from xappt.models.plugins.tool import BaseTool


class BaseInterface(BasePlugin, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self.command_runner = CommandRunner()

        self.on_write_stdout = Callback()
        self.on_write_stderr = Callback()
        self.on_tool_chain_modified = Callback()

        self._current_tool_index: int = -1
        self._tool_chain: list[Type[BaseTool]] = []
        self._current_tool: Optional[BaseTool] = None

        self.tool_data: dict[str: Any] = {}  # tool_data will be sent to both BaseTool.__init__ and BaseTool.execute

    @property
    def current_tool_index(self) -> int:
        return self._current_tool_index

    @property
    def tool_count(self) -> int:
        return len(self._tool_chain)

    def add_tool(self, tool_plugin: Union[str, Type[BaseTool]]):
        if isinstance(tool_plugin, str):
            tool_plugin = xappt.managers.plugin_manager.get_tool_plugin(tool_plugin)
        self._tool_chain.append(tool_plugin)
        self.on_tool_chain_modified.invoke()

    def clear_tool_chain(self):
        self._tool_chain.clear()
        self.on_tool_chain_modified.invoke()

    def get_tool(self, index: int) -> Type[BaseTool]:
        return self._tool_chain[index]

    @abc.abstractmethod
    def run(self, **kwargs) -> int:
        for i, tool_class in enumerate(self._tool_chain):
            self._current_tool_index = i
            self._current_tool = tool_class(interface=self, **self.tool_data)
            result = self.invoke(self._current_tool, **self.tool_data)
            self._current_tool = None
            if result != 0:
                return result
        return 0

    @classmethod
    def collection(cls) -> str:
        return "interface"

    @abc.abstractmethod
    def invoke(self, plugin: BaseTool, **kwargs) -> int:
        pass

    @abc.abstractmethod
    def message(self, message: str):
        pass

    @abc.abstractmethod
    def warning(self, message: str):
        pass

    @abc.abstractmethod
    def error(self, message: str, *, details: Optional[str] = None):
        pass

    @abc.abstractmethod
    def ask(self, message: str) -> bool:
        pass

    @abc.abstractmethod
    def progress_start(self):
        pass

    @abc.abstractmethod
    def progress_update(self, message: str, percent_complete: float):
        pass

    @abc.abstractmethod
    def progress_end(self):
        pass

    def write_stdout(self, text: str):
        self.on_write_stdout.invoke(text)

    def write_stderr(self, text: str):
        self.on_write_stderr.invoke(text)

    def run_subprocess(self, command: Union[bytes, str, Sequence], **kwargs) -> int:
        result = self.command_runner.run(command, stdout_fn=self.write_stdout, stderr_fn=self.write_stderr, **kwargs)
        return result.result

    def abort(self):
        if self._current_tool is not None:
            self._current_tool.abort_requested()



#!/usr/bin/env python3
from __future__ import annotations

import pathlib
import re
import setuptools
import shutil

from typing import Generator, List, Pattern


def build_package_list(base_pkg: str, *, exclude: List[str] = None) -> List[str]:
    if exclude is None:
        exclude = []
    packages = [base_pkg]
    for package in setuptools.find_packages(base_pkg):
        pkg_name = f"{base_pkg}.{package}"
        if pkg_name in exclude:
            continue
        packages.append(pkg_name)
    return packages


def get_version(version_path: pathlib.Path) -> str:
    with version_path.open("r") as fp:
        version_contents = fp.read()

    loc = locals()
    exec(version_contents, {}, loc)
    __version__ = loc['__version__']
    assert __version__ is not None

    return __version__


def update_build(version_path: pathlib.Path, new_build: str):
    version = get_version(version_path)
    with version_path.open("w", newline="\n") as fp:
        fp.write(f'__version__ = "{version}"\n')
        fp.write(f'__build__ = "{new_build}"\n')


def requirements(project_path: pathlib.Path, variation=None) -> Generator[str, None, None]:
    req_file = "requirements"
    if variation is not None:
        req_file = f"requirements-{variation}"
    req_path = project_path.joinpath(f"{req_file}.txt")
    with req_path.open("rb") as fp:
        for line in fp.readlines():
            line = line.decode('utf-8').strip()
            if len(line) == 0:
                continue
            if line.startswith("#"):
                continue
            yield line


def ask(prompt: str) -> bool:
    while True:
        result = input(f"{prompt} (y|n): ").lower()
        if result not in ("y", "n"):
            print(f"Invalid response '{result}', please try again.")
            continue
        if result == "n":
            return False
        return True


def cleanup_build_files(project_root: pathlib.Path) -> bool:
    build_folders = "build", "dist", re.compile(r".*\.egg-info")
    to_clean: list[pathlib.Path] = []
    for item in project_root.iterdir():
        if not item.is_dir():
            continue
        for name_match in build_folders:
            if isinstance(name_match, Pattern):
                re_match = name_match.match(item.name)
                if re_match is not None:
                    to_clean.append(item)
                    break
            else:
                if item.name == name_match:
                    to_clean.append(item)
                    break
    if len(to_clean):
        print("WARNING: The following folders will be permanently deleted:\n")
        for path in to_clean:
            print(f"  {path}")
        print("")
        if ask("Are you sure you want to delete these folders?"):
            for path in to_clean:
                shutil.rmtree(path)
        else:
            if not ask("Proceed anyway?"):
                return False
    return True



import xappt


ITEM_MAP = {
    "first": ("apple", "banana", "cantaloupe"),
    "second": ("red", "green", "blue"),
    "third": ("alpha", "beta", "gamma"),
}


@xappt.register_plugin
class ModifyChoices(xappt.BaseTool):
    integer_choice = xappt.ParamInt(description="Choose an item", choices=list(ITEM_MAP.keys()))
    string_choice = xappt.ParamString(description="Choose an item", choices=ITEM_MAP['first'])

    @classmethod
    def help(cls) -> str:
        return "An example of using callbacks to dynamically modify a parameter's choices"

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def __init__(self, interface: xappt.BaseInterface, **kwargs):
        super().__init__(interface=interface, **kwargs)
        self.integer_choice.on_value_changed.add(self.on_integer_choice_changed)

    def on_integer_choice_changed(self, param: xappt.Parameter):
        item_key_string = param.choices[param.value]
        new_choices = ITEM_MAP[item_key_string]
        self.string_choice.choices = new_choices

    def execute(self, **kwargs) -> int:
        self.interface.message(f"You chose {self.string_choice.value}")
        return 0



from xappt.models.mixins import ConfigMixin
from xappt.utilities.path import user_data_path


class BasePlugin(ConfigMixin):
    def __init__(self):
        super().__init__()
        self._data_dict = {}
        config_file = f"{self.collection()}-{self.name()}.cfg"
        self.config_path = user_data_path().joinpath("xappt").joinpath("plugins").joinpath(config_file)
        self.init_config()

    def init_config(self):
        pass

    @classmethod
    def name(cls) -> str:
        """ Use this to specify the name of your plugin. """
        return cls.__name__.lower()

    @classmethod
    def help(cls) -> str:
        """ Provide a short help string or description. """
        return ""

    @classmethod
    def collection(cls) -> str:
        """ Use this to group your plugin with other like plugins. """
        return ""



from collections import defaultdict
from typing import Optional

import xappt


class TestInterface(xappt.BaseInterface):
    def __init__(self):
        super().__init__()
        self.ask_result: bool = True
        self.messages = defaultdict(list)

    def run(self, **kwargs) -> int:
        return super().run(**kwargs)

    def invoke(self, plugin: xappt.BaseTool, **kwargs) -> int:
        return plugin.execute(**kwargs)

    def message(self, message: str):
        self.messages['message'].append(message)

    def warning(self, message: str):
        self.messages['warning'].append(message)

    def error(self, message: str, *, details: Optional[str] = None):
        self.messages['error'].append((message, details))

    def ask(self, message: str) -> bool:
        return self.ask_result

    def progress_start(self):
        pass

    def progress_update(self, message: str, percent_complete: float):
        pass

    def progress_end(self):
        pass
